/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.text.SimpleDateFormat;
import java.util.Date;

import org.junit.Assert;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAUtensilien;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

/**
 * Simulation eines Lve Datenendgeraetes.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class LveDeSim extends DeSim {

	/**
	 * Fehlercode fuer Negative Quittung
	 */
	private static final int PUFFERINHALT_NICHTVERFUEGBAR = 12;
	/**
	 * Ob fuer LveBetriebsParameter Daten anmeldet ist
	 */
	private boolean anmeldetBetriebsParam;
	/**
	 * Ob fuer LangZeit VerkehrsDaten anmeldet ist
	 */
	private boolean anmeldetLZD;
	/**
	 * Ob fuer LveNegativeQuittung anmeldet ist
	 */
	private boolean anmeldetLveNegQuittung;
	/**
	 * Datenbeschreibung für LVE-LZD
	 */
	private static DataDescription ddLveLzdSend;
	/**
	 * Datenbeschreibung von LVE-Betriesparameter
	 */
	private static DataDescription ddLveParam;
	/**
	 * Datenbeschreibung von LVE-NegativeQuittung
	 */
	private static DataDescription ddLveNegQuit;
	/**
	 * Datenbeschreibung von LVE-NegativeQuittung
	 */
	private static DataDescription dLveArchivParam;
	/**
	 * Der Pid der assoziierten Fahrstreifen
	 */
	private final String fsPid;
	/**
	 * Erzeugungsintervall
	 */
	private final long erzeugungIntervall;
	/**
	 * Der Betriebsparameter mit dem erzeugungsintervall
	 */
	private final long erzeugungIntBetriebParam;
	/**
	 * Ob man negative Quittung senden soll
	 */
	private boolean anfrageNichtVerstanden;

	private final int r = 0;

	/**
	 * Standardkonstruktor
	 *
	 * @param dePid
	 *            Pid des DEs
	 * @param fsPid
	 *            Pid des FS
	 * @param erzeugungsIntervall
	 *            Erzeugungsintervall der LZD
	 */
	public LveDeSim(final String dePid, final String fsPid, final long erzeugungsIntervall) {
		super(dePid);
		this.fsPid = fsPid;
		this.erzeugungIntervall = erzeugungsIntervall * AbfrPufferKonstanten.STUNDE_IN_MS;
		this.erzeugungIntBetriebParam = AbfrPufferKonstanten.convertZeitZurLveZeitCode(this.erzeugungIntervall);
		setLetzterZeitStempellGeschickt(AbfrPufferKonstanten.ZEIT_BEGINN);
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		super.initialize(connection);
		// BetriebsParameter Senden
		LveDeSim.ddLveParam = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_LVEBETRIEBSPARAM,
				AbfrPufferKonstanten.ASP_PARAMETERSOLL, (short) 0);
		DeSim.getDav().subscribeSender(LveDeSim.this, DeSim.getDav().getDataModel().getObject(getDePid()),
				LveDeSim.ddLveParam, SenderRole.source());
		anmeldetBetriebsParam = true;

		// NegativeQuittung Senden
		LveDeSim.ddLveNegQuit = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_LVENEGATIVE_QUITTUNG,
				AbfrPufferKonstanten.ASP_TLS_ANTWORT, (short) 0);
		DeSim.getDav().subscribeSender(this, DeSim.getDav().getDataModel().getObject(getDePid()), LveDeSim.ddLveNegQuit,
				SenderRole.source());
		anmeldetLveNegQuittung = true;

		// Langzeitdaten senden
		if (fsPid != null) {
			LveDeSim.ddLveLzdSend = erzeugeDatenBeschreibung(DUAKonstanten.ATG_LZD, DUAKonstanten.ASP_EXTERNE_ERFASSUNG,
					(short) 0);
			DeSim.getDav().subscribeSender(this, DeSim.getDav().getDataModel().getObject(fsPid), LveDeSim.ddLveLzdSend,
					SenderRole.source());
		}
		try {
			Thread.sleep(100L);
		} catch (final InterruptedException e) {
		}
	}

	/**
	 * Sendet LangZeitDaten Telegrame, wenn keine Daten vorhanden sind, dann
	 * eine Negative Quittung
	 *
	 * @param zeitStempel
	 *            ZeitStemepel des ersten Telegramms
	 * @param anzahl
	 *            Anzahl der Telegramms
	 */
	public void sendeLZD(final long zeitStempel, final long anzahl) {
		if (isOhneAntwort()) {
			return;
		}
		final long zeit = ZeitSimulator.getTimeMillis();

		// ////////////
		// JUnit Tests
		// ////////////
		Assert.assertTrue("LveDe: zeitStempel<zeit :" + new Date(zeitStempel) + "<" + new Date(zeit),
				zeitStempel < zeit);
		// Assert.assertTrue("LveDe: Anzahl " + anzahl, anzahl < (zeit -
		// zeitStempel)/erzeugungIntervall);
		// Assert.assertTrue("LveDe: Anfrage bekommen, obwohl nicht
		// online bin",
		// online && !ohneAntwort);
		// Assert.assertTrue("LveDe: Daten Luecke " + this.dePid + " " +
		// new
		// Date(letzterZeitStempellGeschickt + erzeugungIntervall) + "
		// Anfrage "
		// + new Date(zeitStempel), letzterZeitStempellGeschickt +
		// erzeugungIntervall > zeitStempel);

		// Wenn wir Daten verloren haben dann liefern wir nur die, die vorhanden
		// sind, nicht alle

		final long usedZeitStempel;
		final long usedAnzahl;
		if (getLetzterDatenSatzVorhanden() > zeitStempel) {

			usedZeitStempel = getLetzterDatenSatzVorhanden();
			usedAnzahl = (zeit - usedZeitStempel) / erzeugungIntervall;
		} else {
			usedZeitStempel = zeitStempel;
			usedAnzahl = anzahl;
		}
		if (anfrageNichtVerstanden) {
			sendeNegativeQuittung();
			return;
		}

		Data datensatz = DeSim.getDav()
				.createData(DeSim.getDav().getDataModel().getAttributeGroup(DUAKonstanten.ATG_LZD));
		datensatz = setLZDleer(datensatz);
		datensatz.getTimeValue("T").setMillis(erzeugungIntervall);
		final ResultData resDatei = new ResultData(DeSim.getDav().getDataModel().getObject(fsPid),
				LveDeSim.ddLveLzdSend, usedZeitStempel, datensatz, true);

		int i = 0;
		for (i = 1; i < (usedAnzahl + 1); i++) {
			try {

				resDatei.setDataTime(
						(usedZeitStempel - (usedZeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall));
				if (anmeldetLZD) {
					DeSim.getDav().sendData(resDatei);
				} else if (fsPid != null) {
					DeSim.getDav().sendData(resDatei);
					// dav.subscribeSource(this, resDatei);
					anmeldetLZD = true;
				}
				setLetzterZeitStempellGeschickt(
						(usedZeitStempel - (usedZeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall));
				// System.out.format("%-60s %-60s\n", dePid, "LZD sendet " +
				// AbfPuffKonstanten.datumFormat.format(new Date(zeitStempel -
				// (zeitStempel % erzeugungIntervall)+ i*erzeugungIntervall)) );
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}

		try {
			final ResultData resDatei2 = new ResultData(DeSim.getDav().getDataModel().getObject(fsPid),
					LveDeSim.ddLveLzdSend, usedZeitStempel, null, false, DataState.NO_DATA);
			resDatei2.setDataTime(
					(usedZeitStempel - (usedZeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall) + 150);
			DeSim.getDav().sendData(resDatei2);
			// System.out.format("%-60s %-60s\n", dePid,
			// "LZD Sendung ende: Keine Daten " );
		} catch (final Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Setzt Attribut in Datensatz
	 *
	 * @param attributName
	 *            Name des Attributs
	 * @param wert
	 *            Wert des Attributs
	 * @param datensatz
	 *            Der Datensatz
	 * @return Der veränderte Datensatz
	 */
	private Data setAttribut(final String attributName, final long wert, final Data datensatz) {
		final Data data = datensatz.createModifiableCopy();

		final long usedWert;
		if (attributName.startsWith("v") && (wert >= 255)) {
			usedWert = -1;
		} else {
			usedWert = wert;
		}
		// System.out.println("Name: " + attributName + "wert :" +
		// String.valueOf(wert));
		DUAUtensilien.getAttributDatum(attributName + ".Wert", data).asUnscaledValue().set(usedWert);
		DUAUtensilien.getAttributDatum(attributName + ".Status.Erfassung.NichtErfasst", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.PlFormal.WertMax", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.PlFormal.WertMin", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.PlLogisch.WertMaxLogisch", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.PlLogisch.WertMinLogisch", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.MessWertErsetzung.Implausibel", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Status.MessWertErsetzung.Interpoliert", data).asUnscaledValue()
		.set(DUAKonstanten.NEIN);
		DUAUtensilien.getAttributDatum(attributName + ".Güte.Index", data).asScaledValue().set(1.0);
		DUAUtensilien.getAttributDatum(attributName + ".Güte.Verfahren", data).asUnscaledValue().set(0);

		return datensatz;
	}

	/**
	 * Initialisiert dem datensatz, damit keine undefinierte Werte enthaltet
	 *
	 * @param datensatz
	 *            Datensatz
	 * @return der modifizierte DS
	 */
	private Data setLZDleer(final Data datensatz) {
		// String postfix = new String("GeschwKlasse");
		final String[] praefix = new String[] { "q", "v", "s", "v85" };
		final String[] aName = new String[] { "Kfz", "PkwÄ", "KfzNk", "PkwG", "Pkw", "Krad", "Lfw", "LkwÄ", "PkwA",
				"Lkw", "Bus", "LkwK", "LkwA", "SattelKfz" };

		Data result = datensatz;
		for (final String element : praefix) {
			for (final String element2 : aName) {
				result = setAttribut(element + element2, -1, result);
			}
		}
		return result;
	}

	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		super.parseArguments(argumentList);
	}

	@Override
	public void update(final ResultData[] results) {
		super.update(results);
	}

	/**
	 * Sendet eine Negative Quittung
	 */
	public void sendeNegativeQuittung() {
		final Data data = DeSim.getDav().createData(
				DeSim.getDav().getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_TLS_LVENEGATIVE_QUITTUNG));

		data.getUnscaledValue(AbfrPufferKonstanten.ATR_CODE).set(LveDeSim.PUFFERINHALT_NICHTVERFUEGBAR);
		data.getUnscaledValue(AbfrPufferKonstanten.ATR_HERSTELLER).set(0);
		data.getTextValue(AbfrPufferKonstanten.ATR_EINZELLTELEGRAMM).setText("Nicht verfuegbar");

		try {
			final ResultData resDatei = new ResultData(DeSim.getDav().getDataModel().getObject(getDePid()),
					LveDeSim.ddLveNegQuit, ZeitSimulator.getTimeMillis(), data);
			DeSim.getDav().sendData(resDatei);
			final SimpleDateFormat dateFormat = new SimpleDateFormat(DUAKonstanten.ZEIT_FORMAT_GENAU_STR);
			System.out.format("%-60s %-60s\n", getDePid(),
					"NEGATIVE QUITTUNG sendet " + dateFormat.format(new Date(resDatei.getDataTime())));
		} catch (final Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {

		if (anmeldetBetriebsParam
				&& dataDescription.getAttributeGroup().getPid().equals(AbfrPufferKonstanten.ATG_TLS_LVEBETRIEBSPARAM)
				&& dataDescription.getAspect().getPid().equals(AbfrPufferKonstanten.ASP_PARAMETERSOLL)
				&& (state == ClientSenderInterface.START_SENDING)) {

			final Data data = DeSim.getDav().createData(
					DeSim.getDav().getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_TLS_LVEBETRIEBSPARAM));

			data.getItem(AbfrPufferKonstanten.ATR_INTERVALLDAUER_LZ).asUnscaledValue().set(erzeugungIntBetriebParam); // wert
			// 136
			// beduetet
			// 8
			// Stunden
			data.getItem(AbfrPufferKonstanten.ATR_VERSION_LZD).asUnscaledValue().set(10); // version
			// 10
			data.getItem(AbfrPufferKonstanten.ATR_VERSION_KZD).asUnscaledValue().set(4);
			data.getItem(AbfrPufferKonstanten.ATR_INTERVALLDAUER_KZD).asUnscaledValue().set(20);
			data.getItem(AbfrPufferKonstanten.ATR_ALPHA1).asUnscaledValue().set(128);
			data.getItem(AbfrPufferKonstanten.ATR_ALPHA2).asUnscaledValue().set(128);
			data.getItem(AbfrPufferKonstanten.ATR_LANGE_GW).asUnscaledValue().set(480);
			data.getItem(AbfrPufferKonstanten.ATR_ART_MWB).asUnscaledValue().set(1);
			data.getItem(AbfrPufferKonstanten.ATR_START_MWB).asUnscaledValue().set(100);

			final ResultData resDatei = new ResultData(object, dataDescription, ZeitSimulator.getTimeMillis(), data);
			try {
				DeSim.getDav().sendData(resDatei);
				// System.out.format("%-60s %-60s\n", dePid,
				// "BetrebsParam sendet");
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		} else if (anmeldetBetriebsParam
				&& dataDescription.getAttributeGroup().getPid().equals(AbfrPufferKonstanten.ATG_TLS_LVEBETRIEBSPARAM)
				&& dataDescription.getAspect().getPid().equals(AbfrPufferKonstanten.ASP_PARAMETERSOLL)
				&& (state == ClientSenderInterface.STOP_SENDING)) {
			final ResultData resDatei = new ResultData(object, dataDescription, ZeitSimulator.getTimeMillis(), null);
			try {
				DeSim.getDav().sendData(resDatei);
				// System.out.format("%-60s %-60s\n", dePid,
				// "BetrebsParam sendung ende");
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		} else {
			super.dataRequest(object, dataDescription, state);
		}
	}

	@Override
	public void unsubscribe() {
		if (anmeldetBetriebsParam) {
			DeSim.getDav().unsubscribeSender(LveDeSim.this, DeSim.getDav().getDataModel().getObject(getDePid()),
					LveDeSim.ddLveParam);
		}
		if (anmeldetLZD) {
			DeSim.getDav().unsubscribeSender(LveDeSim.this, DeSim.getDav().getDataModel().getObject(fsPid),
					LveDeSim.ddLveLzdSend);
		}
		if (anmeldetLveNegQuittung) {
			DeSim.getDav().unsubscribeReceiver(this, DeSim.getDav().getDataModel().getObject(getDePid()),
					LveDeSim.ddLveNegQuit);
		}
		super.unsubscribe();
	}

	/**
	 * Liefert den Erzeugeungsintervall in ms
	 *
	 * @return Erzeugungsintervall
	 */
	public long getErzeugungIntervall() {
		return erzeugungIntervall;
	}

	/**
	 * Liefert das PID des Fahrstreifens
	 *
	 * @return PID
	 */
	public String getFsPid() {
		return fsPid;
	}

	/**
	 * Setzt dem Flag, wenn der SystemDe immer mit negativer Quittung antwortet
	 *
	 * @param anfrageNichtVerstanden
	 *            true, wenn negative quitungen geschickt werden sollen
	 */
	public void setAnfrageNichtVerstanden(final boolean anfrageNichtVerstanden) {
		this.anfrageNichtVerstanden = anfrageNichtVerstanden;
	}

	protected boolean isAnfrageNichtVerstanden() {
		return anfrageNichtVerstanden;
	}
}
