/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.dua.ObjektWecker;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Enthaet den Status und andere Informationen zu einem Stuermodul/Station.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class Station {
	/**
	 * Der Logger.
	 */
	public static final Debug LOGGER = Debug.getLogger();
	/**
	 * Das assoziierte SteuerModul.
	 */
	private SystemObject soSteuerModul;
	/**
	 * Das System DE fuer Achslastdaten.
	 */
	private SystemDatenEndgeraet systemAxlDe;
	/**
	 * Das System DE fuer LokaleVerkehrsErfassung Daten.
	 */
	private SystemDatenEndgeraet systemLveDe;
	/**
	 * Liste aller AXL DEs.
	 */
	private final List<DatenEndgeraet> axlDEListe = new ArrayList<>();
	/**
	 * Liste aller Lve DEs.
	 */
	private final List<DatenEndgeraet> lveDEListe = new ArrayList<>();
	/**
	 * Abbildung von Namen von DE zum DE Strukturen.
	 */
	private static Map<String, DatenEndgeraet> mapNameZumDE = new Hashtable<>();
	/**
	 * Ob alle DEs online sein muessen um eine Abfrage zu machen.
	 */
	private static boolean alleMuessenOnlineSein = true;
	/**
	 * Zum sychronisierung der Threads.
	 */
	private final Object mutex;
	/**
	 * Zum timeout intervalle beenden.
	 */
	private static ObjektWecker timeoutWecker = new ObjektWecker();
	/**
	 * Zum periodischen Aufrufen der GeneriereAbrage Methoden.
	 */
	private static AbstraktDeVerwaltung lveDeVerwaltung, axlDeVerwaltung;

	/**
	 * Standard Konstruktor.
	 *
	 * @param steuerModul
	 *            Das SteuerModul das zu die Station gehoert
	 * @param dav
	 *            Verbindung zum DatenVerteiler
	 * @param mutex
	 *            Ein Objekt fuer Synchronizierung
	 * @throws Exception
	 *             Bei Fehlerhaften konfiguration
	 */
	public Station(final SystemObject steuerModul, final ClientDavInterface dav, final Object mutex) throws Exception {
		initialisiere(steuerModul, dav);
		this.mutex = mutex;
	}

	/**
	 * Erfragt, ob ein von den lve DEs flackert.
	 *
	 * @param zeitPunkt
	 *            Zeitstemepl jetzt
	 * @return true, wenn das Lve teil der Station flackert
	 */
	public boolean lveFlackert(final long zeitPunkt) {
		if (Station.alleMuessenOnlineSein) {
			for (final DatenEndgeraet de : lveDEListe) {
				if (de.flackert(zeitPunkt)) {
					return true;
				}
			}
			if ((systemLveDe != null) && systemLveDe.flackert(zeitPunkt)) {
				return true;
			}
			return false;
		} else {
			for (final DatenEndgeraet de : lveDEListe) {
				if (!de.flackert(zeitPunkt)) {
					return false;
				}
			}
			if ((systemLveDe != null) && !systemLveDe.flackert(zeitPunkt)) {
				return false;
			}
			return true;
		}
	}

	/**
	 * Erfragt, ob ein von den axl DEs flackert.
	 *
	 * @param zeitPunkt
	 *            Zeitstemepl jetzt
	 * @return true, wenn das Axl teil der Station flackert
	 */
	public boolean axlFlackert(final long zeitPunkt) {
		if (Station.alleMuessenOnlineSein) {
			for (final DatenEndgeraet de : axlDEListe) {
				if (de.flackert(zeitPunkt)) {
					return true;
				}
			}
			if ((systemAxlDe != null) && systemAxlDe.flackert(zeitPunkt)) {
				return true;
			}
			return false;
		} else {
			for (final DatenEndgeraet de : axlDEListe) {
				if (!de.flackert(zeitPunkt)) {
					return false;
				}
			}
			if ((systemAxlDe != null) && !systemAxlDe.flackert(zeitPunkt)) {
				return false;
			}
			return true;
		}
	}

	/**
	 * Wenn das SystemObjekt kein Konfigurationsobjekt ist, dann wird dieses
	 * Exception geworfen.
	 */
	public class KeinKonfigurationsObjektException extends Exception {

		private static final long serialVersionUID = 1L;

		KeinKonfigurationsObjektException(final String s) {
			super(s);
		}
	}

	/**
	 * Initialisiert die Klasse, aus dem SystemObject.
	 *
	 * @param steuerModul
	 *            Steuermodul SystemObjekt
	 * @param dav
	 *            Datenvertailerverbindung
	 * @throws KeinKonfigurationsObjektException
	 *             wird geworfen, wenn das soSteuermodul kein Konf. Objekt ist
	 */
	public void initialisiere(final SystemObject steuerModul, final ClientDavInterface dav)
			throws KeinKonfigurationsObjektException {

		this.soSteuerModul = steuerModul;

		if (!(steuerModul instanceof ConfigurationObject)) {
			throw new KeinKonfigurationsObjektException(
					"Objekt " + steuerModul.getPid() + " ist kein Konfigurationsobjekt");
		}

		final ConfigurationObject confObj = (ConfigurationObject) steuerModul;
		final ObjectSet eakMenge = confObj.getObjectSet(AbfrPufferKonstanten.ATR_EAK);
		if (eakMenge == null) {
			Station.LOGGER.warning("SteuerModul " + steuerModul.getPid() + " enthaelt keine Menge von EAKs");
			return;
		}

		final List<SystemObject> eakListe = eakMenge.getElements();
		for (final SystemObject eakObj : eakListe) {
			if (eakObj.isValid()) {
				if (!(eakObj instanceof ConfigurationObject)) {
					throw new KeinKonfigurationsObjektException(eakObj + " ist kein Konfigurationsobjekt");
				}

				final ConfigurationObject eakConfObj = (ConfigurationObject) eakObj;
				final ObjectSet deMenge = eakConfObj.getObjectSet(AbfrPufferKonstanten.ATR_DE);
				if (deMenge == null) {
					Station.LOGGER.warning("EAK " + eakConfObj.getPid() + " enthaelt keine Menge von DEs");
					return;
				}
				final List<SystemObject> deListe = deMenge.getElements();
				for (final SystemObject deObj : deListe) {
					if (deObj.isValid()) {

						if (deObj.isOfType(AbfrPufferKonstanten.TYP_DE_LVE)) {
							final Data datei = deObj.getConfigurationData(
									dav.getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_DE),
									dav.getDataModel().getAspect(AbfrPufferKonstanten.ASP_EIGENSCHAFTEN));

							final int deKanal = Integer
									.parseInt(datei.getItem(AbfrPufferKonstanten.ATR_DE_KANAL).asTextValue().getText());
							// DEs mit kanal 255 sind SYS-DEs
							if (deKanal == 255) {
								systemLveDe = new SystemDatenEndgeraet(deObj, this);
								Station.mapNameZumDE.put(deObj.getPid(), systemLveDe);
							} else {
								final String clusterStateName = datei.getUnscaledValue("Cluster").getState().getName();
								if ("Nein".equals(clusterStateName)) {
									final DatenEndgeraet de = new DatenEndgeraet(deObj, this);
									lveDEListe.add(de);
									Station.mapNameZumDE.put(deObj.getPid(), de);
								}
							}
						} else if (deObj.isOfType(AbfrPufferKonstanten.TYP_DE_AXL)) {
							final Data datei = deObj.getConfigurationData(
									dav.getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_DE),
									dav.getDataModel().getAspect(AbfrPufferKonstanten.ASP_EIGENSCHAFTEN));

							final int deKanal = Integer
									.parseInt(datei.getItem(AbfrPufferKonstanten.ATR_DE_KANAL).asTextValue().getText());
							if (deKanal == 255) {
								systemAxlDe = new SystemDatenEndgeraet(deObj, this);
								Station.mapNameZumDE.put(deObj.getPid(), systemAxlDe);
							} else {
								final DatenEndgeraet de = new DatenEndgeraet(deObj, this);
								axlDEListe.add(de);
								Station.mapNameZumDE.put(deObj.getPid(), de);
							}
						}
					}
				}

			}
		}

		// ###################################################################################################
		/*
		 * ES GIBT KEIN SYS-DE(ODER MIT K255) IN DER KONFIGURATIONMENGE, SYSTEM
		 * DE NACH NAMEN GESUCHT (NICHT ZUVERLAESSIG)!!!
		 */
		// ###################################################################################################
		if (!lveDEListe.isEmpty() && (systemLveDe == null)) {
			String pid = lveDEListe.get(0).getSoDE().getPid();
			pid = pid.replaceAll("DeLve[\\d]*", "DeSammelkanal_1");
			systemLveDe = new SystemDatenEndgeraet(dav.getDataModel().getObject(pid), this);
			Station.mapNameZumDE.put(pid, systemLveDe);
			Station.LOGGER.config("System Datenendgeraet nach Namen erkannt " + pid);
		}
		if (!axlDEListe.isEmpty() && (systemAxlDe == null)) {
			String pid = axlDEListe.get(0).getSoDE().getPid();
			pid = pid.replaceAll("DeAxl[\\d]*", "DeSammelkanal_1");
			systemAxlDe = new SystemDatenEndgeraet(dav.getDataModel().getObject(pid), this);
			Station.mapNameZumDE.put(pid, systemAxlDe);
			Station.LOGGER.config("System Datenendgeraet nach Namen erkannt " + pid);
		}
	}

	@Override
	public String toString() {
		final long jetzt = Station.lveDeVerwaltung.currentTimeMillis();
		final StringBuffer s = new StringBuffer();
		String s2;
		if (this.systemLveDe != null) {
			s2 = systemLveDe.toString();
		} else {
			s2 = "null";
		}
		s.append(String.format("STEUER MODUL: %-60s\nSYSTEM Lve DE: %-60s", soSteuerModul.getPid(), s2));

		for (final DatenEndgeraet de : lveDEListe) {
			s.append(de.toString(jetzt));
		}

		if (this.systemAxlDe != null) {
			s2 = systemAxlDe.toString();
		} else {
			s2 = "null";
		}
		s.append(String.format("STEUER MODUL: %-60s\nSYSTEM Axl DE: %-60s", soSteuerModul.getPid(), s2));

		for (final DatenEndgeraet de : axlDEListe) {
			s.append(de.toString(jetzt));
		}

		return s.toString();
	}

	/**
	 * Liefert eine Referenz auf eine DatenEndgeraet Instanz des DEs vom Pid.
	 *
	 * @param pid
	 *            Pid des DEs
	 * @return DatenEndgeraet
	 */
	public static DatenEndgeraet getDatenEndgeraet(final String pid) {
		return Station.mapNameZumDE.get(pid);
	}

	/**
	 * Erfragt, ob das De mit Pid in der Abbildung auf Datenendgeraetinstanzen
	 * enthalten ist.
	 *
	 * @param pid
	 *            Pid des DEs
	 * @return DatenEndgeraet
	 */
	public static boolean enthaeltEndgeraet(final String pid) {
		return Station.mapNameZumDE.containsKey(pid);
	}

	/**
	 * Liefert das System Axl De.
	 *
	 * @return the SystemAxlDe
	 */
	public SystemDatenEndgeraet getSystemAxlDe() {
		return systemAxlDe;
	}

	/**
	 * Liefert das System Lve De.
	 *
	 * @return SystemLveDe
	 */
	public SystemDatenEndgeraet getSystemLveDe() {
		return systemLveDe;
	}

	/**
	 * Erfragt die Liste der Axl DEs.
	 *
	 * @return Axl DE Liste
	 */
	public Collection<DatenEndgeraet> getAxlDeListe() {
		return axlDEListe;
	}

	/**
	 * Erfragt die Liste der Lve DEs.
	 *
	 * @return Lve DE Liste
	 */
	public Collection<DatenEndgeraet> getLveDeListe() {
		return lveDEListe;
	}

	/**
	 * Ergibt die Menge der Axl DEs, eine Kopie der Liste.
	 *
	 * @return Menge der Axl DEs
	 */
	public Collection<SystemObject> getAlleAxlDEs() {
		final Collection<SystemObject> axl = new LinkedList<>();
		for (final DatenEndgeraet de : axlDEListe) {
			axl.add(de.getSoDE());
		}
		return axl;
	}

	/**
	 * Erfragt die Menge der Lve DEs, eine Kopie der Liste.
	 *
	 * @return Mnege der Lve DEs
	 */
	public final Collection<SystemObject> getAlleLveDEs() {
		final Collection<SystemObject> lve = new LinkedList<>();
		for (final DatenEndgeraet de : lveDEListe) {
			lve.add(de.getSoDE());
		}
		return lve;
	}

	/**
	 * Die Aelteste Lve Zeitstempel von allen DEs die parametriert und online
	 * sind.
	 *
	 * @return Zeitstempel
	 */
	public long getAeltesteLveZeitStempel() {
		final long jetzt = Station.lveDeVerwaltung.currentTimeMillis();
		long zeit = jetzt;

		for (final DatenEndgeraet de : lveDEListe) {

			if (de.istParametriert() && de.isOnline() && (de.getLetzterDatensatzZeitStempel() < zeit)) {
				zeit = de.getLetzterDatensatzZeitStempel();
			}
		}
		return zeit;
	}

	/**
	 * Die Aelteste Axl Zeitstempel von allen DEs die parametriert und online
	 * sind.
	 *
	 * @return Zeitstempel
	 */
	public long getAeltestenAxlZeitStempel() {
		final long jetzt = Station.axlDeVerwaltung.currentTimeMillis();
		long zeit = jetzt;

		for (final DatenEndgeraet de : axlDEListe) {
			if (de.istParametriert() && de.isOnline() && (de.getLetzterDatensatzZeitStempel() < zeit)) {
				zeit = de.getLetzterDatensatzZeitStempel();
			}
		}
		return zeit;
	}

	/**
	 * Die Maximale Anzahl der fehlenden Datensaetzen von allen Lve DEs.
	 *
	 * @param zeitstempel
	 *            Zeitintervall obere Grenze
	 * @return Anzahl der Datensaetze vom letzten Zeitstempel bis obere Grenze
	 */
	public long getLveMaxAnzahlFehlendenDatenSaetzen(final long zeitstempel) {
		long anzahl = 0;
		for (final DatenEndgeraet de : lveDEListe) {

			if (de.istParametriert() && de.isOnline()) {
				// minus 1, weil Zeitstempel immer Beginn ist
				final long n = ((zeitstempel - de.getLetzterDatensatzZeitStempel()) / de.getErzeugungsIntervall()) - 1;
				if (n > anzahl) {
					anzahl = n;
				}
			}
		}
		return anzahl;
	}

	/**
	 * Die Maximale Anzahl der fehlenden Datensaetzen von allen Axl DEs.
	 *
	 * @param zeitstempel
	 *            Zeitintervall obere Grenze
	 * @return Anzahl der Datensaetze vom letzten Zeitstempel bis obere Grenze
	 */
	public long getAxlMaxAnzahlFehlendenDatenSaetzen(final long zeitstempel) {
		long anzahl = 0;
		for (final DatenEndgeraet de : axlDEListe) {

			if (de.istParametriert() && de.isOnline()) {
				// minus 1, weil Zeitstempel immer Beginn ist
				final long n = ((zeitstempel - de.getLetzterDatensatzZeitStempel()) / de.getErzeugungsIntervall()) - 1;
				if (n > anzahl) {
					anzahl = n;
				}
			}
		}
		return anzahl;
	}

	/**
	 * Erfragt den minimalen Erzeugungsintervall der LZD aller Lve DEs.
	 *
	 * @return ZeitIntervall
	 */
	public long getLveLZDErzeugungsIntervallMinimum() {
		long zeit = AbfrPufferKonstanten.INTERVALL_MAX;

		for (final DatenEndgeraet de : lveDEListe) {
			if (de.istParametriert()) {
				final long temp = de.getErzeugungsIntervall();
				if (temp < zeit) {
					zeit = temp;
				}
			}
		}

		return zeit;
	}

	/**
	 * Erfragt den minimalen Erzeugungsintervall der LZD aller Axl DEs.
	 *
	 * @return ZeitIntervall
	 */
	public long getAxlDatenErzeugungsIntervallMinimum() {
		long zeit = AbfrPufferKonstanten.INTERVALL_MAX;

		for (final DatenEndgeraet de : axlDEListe) {
			if (de.istParametriert()) {
				final long temp = de.getErzeugungsIntervall();
				if (temp < zeit) {
					zeit = temp;
				}
			}
		}

		return zeit;
	}

	/**
	 * Erfragt, ob alle Lve Des parametriert sind.
	 *
	 * @return true, wenn alle Lve DEs parametriert sind
	 */
	public boolean istLveParametriert() {
		for (final DatenEndgeraet de : lveDEListe) {
			if (!de.istParametriert()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Erfragt ob alle Axl Des online sind.
	 *
	 * @return true, wenn Axl DEs der Statuib online sind
	 */
	public boolean istAxlOnline() {
		if (Station.alleMuessenOnlineSein) {
			for (final DatenEndgeraet de : axlDEListe) {
				if (!de.isOnline()) {
					return false;
				}
			}
			return true;
		} else {
			for (final DatenEndgeraet de : axlDEListe) {
				if (de.isOnline()) {
					return true;
				}
			}
			return false;
		}
	}

	/**
	 * Erfragt ob alle Axl Des parametriert sind.
	 *
	 * @return true, wenn axl DEs parametriert sind
	 */
	public boolean istAxlParametriert() {
		for (final DatenEndgeraet de : axlDEListe) {
			if (!de.istParametriert()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Erfragt ob alle Lve Des online sind.
	 *
	 * @return true, wenn Lve DEs der Statuib online sind
	 */
	public boolean istLveOnline() {
		if (Station.alleMuessenOnlineSein) {
			for (final DatenEndgeraet de : lveDEListe) {
				if (!de.isOnline()) {
					return false;
				}
			}
			return true;
		} else {
			for (final DatenEndgeraet de : lveDEListe) {
				if (de.isOnline()) {
					return true;
				}
			}
			return false;
		}
	}

	/**
	 * Ueberprueft, ob alle DEs datensaetze bis zum Zeitpunkt bekommen haben.
	 *
	 * @param zeitpunkt
	 *            Zeitstempel
	 * @return true, wenn alle Datensaetze eingekommen sind
	 */
	public boolean sindAlleLveDatenSaetzeGekommen(final long zeitpunkt) {
		for (final DatenEndgeraet de : lveDEListe) {
			if (!de.istParametriert() || !de.isOnline()) {
				continue;
			}

			if ((zeitpunkt - de.getLetzterDatensatzZeitStempel()) < de.getErzeugungsIntervall()) {
				// Es kommen nicht mehrere weil Zeitstempel aktuell ist
				if (de.getAnzahlErwarteteDatenSaetze() > 0) {
					de.setAnzahlErwarteteDatenSaetze(0);
				}
			} else {
				// ZeitStempel nicht aktuell und nicht alle sind eingekommen
				if (de.getAnzahlErwarteteDatenSaetze() > 0) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Ueberprueft, ob alle Axl DEs alle Datensaetze bis zum Zeitpunkt bekommen
	 * haben.
	 *
	 * @param zeitpunkt
	 *            ZeitStempel
	 * @return true, wenn alle Datensaetze eingekommen sind
	 */
	public boolean sindAlleAxlDatenSaetzeGekommen(final long zeitpunkt) {
		for (final DatenEndgeraet de : axlDEListe) {
			if (!de.istParametriert() || !de.isOnline()) {
				continue;
			}

			if ((zeitpunkt - de.getLetzterDatensatzZeitStempel()) < de.getErzeugungsIntervall()) {
				// Es kommen nicht mehrere weil Zeitstempel aktuell ist
				if (de.getAnzahlErwarteteDatenSaetze() > 0) {
					de.setAnzahlErwarteteDatenSaetze(0);
				}
			} else {
				// ZeitStempel nicht aktuell und nicht alle sind eingekommen
				if (de.getAnzahlErwarteteDatenSaetze() > 0) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Stellt die Anzahl der Erwarteten Datensaetzen ein.
	 *
	 * @param anzahl
	 *            Anzahl
	 */
	public void setErwarteLveDatensaetze(final long anzahl) {
		for (final DatenEndgeraet de : lveDEListe) {
			de.setAnzahlErwarteteDatenSaetze(anzahl);
		}
	}

	/**
	 * Setzt die Anzahl der Erwarteten Datensaetzen.
	 *
	 * @param anzahl
	 *            Anzahl
	 */
	public void setErwarteAxlDatensaetze(final long anzahl) {
		for (final DatenEndgeraet de : axlDEListe) {
			de.setAnzahlErwarteteDatenSaetze(anzahl);
		}
	}

	/**
	 * Statische Metode, erfragt die Menge aller System Lve DEs aus der Liste
	 * der steuerModule.
	 *
	 * @param steuerModule
	 *            Liste der SM
	 * @return Menge aller System Lve DEs die zu SM in der Liste gehoeren
	 */
	public static Collection<SystemObject> getAlleLveSystemDEs(final Collection<Station> steuerModule) {
		final Collection<SystemObject> lveDEs = new LinkedList<>();
		for (final Station sm : steuerModule) {
			final SystemDatenEndgeraet sde = sm.getSystemLveDe();
			if (sde != null) {
				final SystemObject so = sde.getSoDE();
				if (so != null) {
					lveDEs.add(so);
				}
			}
		}
		return lveDEs;
	}

	/**
	 * Statische Metode, erfragt die Menge aller Lve DEs aus der Liste der
	 * steuerModule.
	 *
	 * @param steuerModule
	 *            Liste der SM
	 * @return Menge aller Lve DEs die zu SM in der Liste gehoeren
	 */
	public static Collection<SystemObject> getAlleLveDEs(final Collection<Station> steuerModule) {
		final Collection<SystemObject> lveDEs = new LinkedList<>();
		for (final Station sm : steuerModule) {
			final Collection<SystemObject> des = sm.getAlleLveDEs();
			lveDEs.addAll(des);
		}
		return lveDEs;
	}

	/**
	 * Statische Metode, erfragt die Menge aller System Axl DEs aus der Liste
	 * der steuerModule.
	 *
	 * @param steuerModule
	 *            Liste der SM
	 * @return Menge aller System Axl DEs die zu SM in der Liste gehoeren
	 */
	public static Collection<SystemObject> getAlleAxlSystemDEs(final Collection<Station> steuerModule) {
		final Collection<SystemObject> axlDEs = new LinkedList<>();
		for (final Station sm : steuerModule) {
			final SystemDatenEndgeraet sde = sm.getSystemAxlDe();
			if (sde != null) {
				final SystemObject so = sde.getSoDE();
				if (so != null) {
					axlDEs.add(so);
				}
			}
		}
		return axlDEs;
	}

	/**
	 * Statische Metode, erfragt die Menge aller Axl DEs aus der Liste der
	 * steuerModule.
	 *
	 * @param steuerModule
	 *            Liste der SM
	 * @return Menge aller Axl DEs die zu SM in der Liste gehoeren
	 */
	public static Collection<SystemObject> getAlleAxlDEs(final Collection<Station> steuerModule) {
		final Collection<SystemObject> axlDEs = new LinkedList<>();
		for (final Station sm : steuerModule) {
			final Collection<SystemObject> des = sm.getAlleAxlDEs();
			axlDEs.addAll(des);
		}
		return axlDEs;
	}

	/**
	 * Statische Metode, erfragt die Menge aller DEs aus der Liste der
	 * steuerModule.
	 *
	 * @param steuerModule
	 *            Liste der SM
	 * @return Menge aller DEs die zu SM in der Liste gehoeren
	 */
	public static Collection<SystemObject> getAlleDEs(final Collection<Station> steuerModule) {
		final Collection<SystemObject> alleDEs = new LinkedList<>();

		alleDEs.addAll(Station.getAlleAxlDEs(steuerModule));
		alleDEs.addAll(Station.getAlleLveDEs(steuerModule));

		return alleDEs;
	}

	/**
	 * Erfragt ob alle DEs online sein muessen, um eine Anfrage zu machen
	 * koennen.
	 *
	 * @return true, wenn eine Abfrage nur dann moeglich ist, wenn alle online
	 *         sind
	 */
	public static boolean isAlleMuessenOnlineSein() {
		return Station.alleMuessenOnlineSein;
	}

	/**
	 * Setzt die statische vairable die bestimmt, ob alle DEs online sein
	 * muessen, um eine Anfrage zu machen koennen.
	 *
	 * @param alleMuessenOnlineSein
	 *            true, wenn eine Abfrage nur dann moeglich ist, wenn alle
	 *            online sind
	 */
	public static void setAlleMuessenOnlineSein(final boolean alleMuessenOnlineSein) {
		Station.alleMuessenOnlineSein = alleMuessenOnlineSein;
	}

	/**
	 * Liefert den Mutex, der zum DatenZugriff dient.
	 *
	 * @return Mutex
	 */
	public Object getMutex() {
		return mutex;
	}

	/**
	 * Bricht einen Timeout ab, wenn alle Lve Datensaetze schon angekommen sind.
	 */
	public void lveTimeoutAbbrechen() {
		Station.timeoutWecker.setWecker(this.systemLveDe, ObjektWecker.AUS);
	}

	/**
	 * Bricht einen Timeout ab, wenn alle Axl Datensaetze schon angekommen sind.
	 */
	public void axlTimeoutAbbrechen() {
		Station.timeoutWecker.setWecker(this.systemAxlDe, ObjektWecker.AUS);
	}

	/**
	 * Setzt den Timeout ein einer Anfrage der Lve Daten ein.
	 *
	 * @param delay
	 *            Laenge des Timeouts
	 */
	public void setLveTimeoutWecker(final long delay) {

		if (this.systemLveDe == null) {
			return;
		}

		final long usedDelay = Station.lveDeVerwaltung.convertTimeoutIntervall(delay);

		if (usedDelay > Station.lveDeVerwaltung.getHauptModul().getAbrufIntervallLve()) {
			Station.LOGGER.error("Lve Timeout laenger als Abrufintervall: Timeout:" + usedDelay + "ms Abrufintervall:"
					+ Station.lveDeVerwaltung.getHauptModul().getAbrufIntervallLve() + "ms");
		}
		final long jetzt = System.currentTimeMillis();
		Station.timeoutWecker.setWecker(this.systemLveDe, jetzt + usedDelay);
	}

	/**
	 * Setzt den Timeout nach einer Anfrage der Axl Daten ein.
	 *
	 * @param delay
	 *            Laenge des Timeouts
	 */
	public void setAxlTimeoutWecker(final long delay) {

		if (this.systemAxlDe == null) {
			return;
		}
		final long usedDelay = Station.lveDeVerwaltung.convertTimeoutIntervall(delay);

		if (usedDelay > Station.axlDeVerwaltung.getHauptModul().getAbrufIntervallAxl()) {
			Station.LOGGER.error("Axl Timeout laenger als Abrufintervall: Timeout:" + usedDelay + "ms Abrufintervall:"
					+ Station.axlDeVerwaltung.getHauptModul().getAbrufIntervallAxl() + "ms");
		}
		final long jetzt = System.currentTimeMillis();
		Station.timeoutWecker.setWecker(this.systemAxlDe, jetzt + usedDelay);
	}

	/**
	 * erfrag dem Wecker fuer Timeouts.
	 *
	 * @return Wecker
	 */
	public ObjektWecker getWecker() {
		return Station.timeoutWecker;
	}

	/**
	 * Erfragt die LveDeVerwaltung Modul Instanz.
	 *
	 * @return LveDeVerwaltung Klasse
	 */
	public static AbstraktDeVerwaltung getLveDeVerwaltung() {
		return Station.lveDeVerwaltung;
	}

	/**
	 * Setzt die lveVerwaltungmodul Instanz.
	 *
	 * @param lveDeVerwaltung
	 *            LveDeVerwaltung Klasse
	 */
	public static void setLveDeVerwaltung(final AbstraktDeVerwaltung lveDeVerwaltung) {
		Station.lveDeVerwaltung = lveDeVerwaltung;
	}

	/**
	 * Erfragt die AxlDeVerwaltung Modul Instanz.
	 *
	 * @return AxlDeVerwaltung Klasse
	 */
	public static AbstraktDeVerwaltung getAxlDeVerwaltung() {
		return Station.axlDeVerwaltung;
	}

	/**
	 * Setzt die axlVerwaltungmodul Instanz.
	 *
	 * @param axlDeVerwaltung
	 *            AxlDeVerwaltung Klasse
	 */
	public static void setAxlDeVerwaltung(final AbstraktDeVerwaltung axlDeVerwaltung) {
		Station.axlDeVerwaltung = axlDeVerwaltung;
	}

	/**
	 * Ermittelt dem Hauptmodul dass, die Abfrage beendet ist.
	 *
	 * @param timeout
	 *            bestimmt, ob nach dem Ablauf eines definierten Timeout
	 *            abgebrochen werden soll.
	 */
	public void abfrageEnde(final boolean timeout) {
		Station.lveDeVerwaltung.abfrageEnde(this, timeout);
	}

	/**
	 * Erfragt nach dem Pid des SteuerModuls.
	 *
	 * @return Pid
	 */
	public String getPid() {
		return soSteuerModul.getPid();
	}

	/**
	 * Loescht alle Flags ueber Negativen Quittungen der einzelnen DEs.
	 */
	public void resetAxlNegativeQuittungen() {
		for (final DatenEndgeraet de : axlDEListe) {
			de.setAntwortetNegQuitt(false);
		}
	}

	/**
	 * Loescht alle Flags ueber Negativen Quittungen der einzelnen DEs.
	 */
	public void resetLveNegativeQuittungen() {
		for (final DatenEndgeraet de : lveDEListe) {
			de.setAntwortetNegQuitt(false);
		}
	}

	/**
	 * Erfragt ob ein DE mit Negativen Quittung geantwortet hat.
	 *
	 * @return <code>true</code> wenn negative Quittung gekommen ist
	 */
	public boolean antwortLveNegQuittBekommen() {
		for (final DatenEndgeraet de : lveDEListe) {
			if (de.antwortetNegQuitt()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Erfragt ob ein DE mit Negativen Quittung geantwortet hat.
	 *
	 * @return <code>true</code> wenn negative Quittung gekommen ist
	 */
	public boolean antwortAxlNegQuittBekommen() {
		for (final DatenEndgeraet de : axlDEListe) {
			if (de.antwortetNegQuitt()) {
				return true;
			}
		}
		return false;
	}
}
