/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

/**
 * Simulation eines Steuermoduls Verwaltet seine Datenendgeraete, und eine Map,
 * die Pids von DEs auf Systemobjekte abbildet
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class SteuerModulSim {
	/**
	 * Bildet Pids von DEs auf SystemObjekte ab
	 */
	private static Map<String, DeSim> mapPidDe = new HashMap<>();
	/**
	 * Liste aller Lve DEs des SteuerModuls
	 */
	private Collection<LveDeSim> lveDeList = new LinkedList<>();
	/**
	 * System De des Steuermoduls fuer Lve Des
	 */
	private LveDeKanal255Sim sysLveDe;
	/**
	 * Liste aller Axl DEs des SteuerModuls
	 */
	private Collection<AxlDeSim> axlDeList = new LinkedList<>();
	/**
	 * System De des Steuermoduls fuer Axl Des
	 */
	private AxlDeKanal255Sim sysAxlDe;
	/**
	 * Ob man fuer die Archivparametrierung als Sender angemeldet ist
	 */
	private static boolean achivParametrierungAngemeldet;

	/**
	 * Erfragt, ob man fuer die Archivparametrierung angemeldet ist
	 *
	 * @return true, wenn man angemeldet ist
	 */
	public static boolean isAchivParametrierungAngemeldet() {
		return SteuerModulSim.achivParametrierungAngemeldet;
	}

	/**
	 * Setzt den Zeitstempel des aeltesten Datensatzes, dem man noch liefern
	 * kann, fuer Axl Daten
	 *
	 * @param smListe
	 *            Liste der Steuermodule
	 * @param zeitStemepel
	 *            zeitStempel
	 */
	public static void setAeltesterAxlZeitStempellVorhanden(final Collection<SteuerModulSim> smListe,
			final long zeitStemepel) {
		for (final SteuerModulSim sm : smListe) {
			sm.setAeltesterAxlZeitStempellVorhanden(zeitStemepel);
		}
	}

	/**
	 * Setzt den Zeitstempel des aeltesten Datensatzes, dem man noch liefern
	 * kann, fuer Lve Daten
	 *
	 * @param smListe
	 *            Liste der Steuermodule
	 * @param zeitStempel
	 *            zeitStempel
	 */
	public static void setAeltesterLveZeitStempellVorhanden(final Collection<SteuerModulSim> smListe,
			final long zeitStempel) {
		for (final SteuerModulSim sm : smListe) {
			sm.setAeltesterLveZeitStempellVorhanden(zeitStempel);
		}
	}

	/**
	 * Setzt den Zeitstempel des aeltesten Datensatzes, dem man noch liefern
	 * kann, fuer Axl Daten
	 *
	 * @param zeitStempel
	 *            zeitStempel
	 */
	public void setAeltesterAxlZeitStempellVorhanden(final long zeitStempel) {
		for (final DeSim de : axlDeList) {
			de.setAeltesterZeitStempellVorhanden(zeitStempel);
		}
	}

	/**
	 * Setzt den Zeitstempel des aeltesten Datensatzes, dem man noch liefern
	 * kann, fuer Lve Daten
	 *
	 * @param zeitStempel
	 *            zeitStempel
	 */
	public void setAeltesterLveZeitStempellVorhanden(final long zeitStempel) {
		for (final DeSim de : lveDeList) {
			de.setAeltesterZeitStempellVorhanden(zeitStempel);
		}
	}

	/**
	 * Setzt den Flag, der bestimmt, ob man als sender fuer Archivparametrierung
	 * angemeldet ist
	 *
	 * @param achivParametrierungAngemeldet
	 *            true, wenn angemeldet
	 */
	public static void setAchivParametrierungAngemeldet(final boolean achivParametrierungAngemeldet) {
		SteuerModulSim.achivParametrierungAngemeldet = achivParametrierungAngemeldet;
	}

	/**
	 * Standardkonstruktor
	 *
	 * @param listAxlDePid
	 *            Liste der Datenendgeraete und Achslastmessstellenpid-Paare
	 * @param sysAxlDePid
	 *            System Axl Datenendgeraet
	 * @param listLveDePid
	 *            Liste der Datenendgeraete und Fahrstreifen-Paare
	 * @param sysLveDePid
	 *            System Lve Datenendgeraet
	 */
	public SteuerModulSim(final Collection<DePid> listAxlDePid, final String sysAxlDePid,
			final Collection<DePid> listLveDePid, final String sysLveDePid) {
		this.createAxlSM(listAxlDePid, sysAxlDePid);
		this.createLveSM(listLveDePid, sysLveDePid);
	}

	/**
	 * Erzeugt den Model des SM fuer Lve Daten, erzeugt Instanzen der DEs
	 *
	 * @param listDePid
	 *            Liste der DEs
	 * @param sysDePid
	 *            System DE
	 */
	public void createLveSM(final Collection<DePid> listDePid, final String sysDePid) {
		if ((listDePid == null) || (sysDePid == null)) {
			return;
		}
		lveDeList = new LinkedList<>();
		for (final DePid dp : listDePid) {
			final LveDeSim sim = new LveDeSim(dp.getDePid(), dp.getFsPid(), dp.getIntervall());
			lveDeList.add(sim);
			SteuerModulSim.mapPidDe.put(dp.getDePid(), sim);
		}
		sysLveDe = new LveDeKanal255Sim(sysDePid, null, lveDeList);
		SteuerModulSim.mapPidDe.put(sysDePid, sysLveDe);
	}

	/**
	 * Erzeugt den Model des SM fuer Axl Daten, erzeugt Instanzen der DEs
	 *
	 * @param listDePid
	 *            Liste der DEs
	 * @param sysDePid
	 *            System DE
	 */
	public void createAxlSM(final Collection<DePid> listDePid, final String sysDePid) {
		if ((listDePid == null) || (sysDePid == null)) {
			return;
		}
		axlDeList = new LinkedList<>();
		for (final DePid dp : listDePid) {
			final AxlDeSim sim = new AxlDeSim(dp.getDePid(), dp.getFsPid(), dp.getIntervall());
			axlDeList.add(sim);
			SteuerModulSim.mapPidDe.put(dp.getDePid(), sim);
		}
		sysAxlDe = new AxlDeKanal255Sim(sysDePid, null, axlDeList);
		SteuerModulSim.mapPidDe.put(sysDePid, sysAxlDe);
	}

	/**
	 * Initialisiert die DEs
	 *
	 * @param connection
	 *            Verbindung zum DAV
	 * @throws Exception
	 *             Exception beim Fehler
	 */
	public void initialize(final ClientDavInterface connection) throws Exception {
		for (final DeSim de : axlDeList) {
			de.initialize(connection);
		}

		for (final DeSim de : lveDeList) {
			de.initialize(connection);
		}

		if (sysLveDe != null) {
			sysLveDe.initialize(connection);
		}

		if (sysAxlDe != null) {
			sysAxlDe.initialize(connection);
		}
	}

	/**
	 * Bearbeitet die Argumente
	 *
	 * @param argumentList
	 *            Liste der Argumene
	 * @throws Exception
	 *             Exception beim Fehler
	 */
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		for (final DeSim de : axlDeList) {
			final ArgumentList argList = new ArgumentList(argumentList.getArgumentStrings());
			de.parseArguments(argList);
		}
		for (final DeSim de : lveDeList) {
			final ArgumentList argList = new ArgumentList(argumentList.getArgumentStrings());
			de.parseArguments(argList);
		}

		if (sysLveDe != null) {
			sysLveDe.parseArguments(argumentList);
		}

		if (sysAxlDe != null) {
			sysAxlDe.parseArguments(argumentList);
		}
	}

	/**
	 * Abmeldung am Ende der Simulation
	 */
	public void unsubscribe() {
		for (final DeSim de : axlDeList) {
			de.unsubscribe();
		}
		for (final DeSim de : lveDeList) {
			de.unsubscribe();
		}

		if (sysLveDe != null) {
			sysLveDe.unsubscribe();
		}

		if (sysAxlDe != null) {
			sysAxlDe.unsubscribe();
		}
	}

	/**
	 * Liefert das SystemObjekt de aus dem Pid
	 *
	 * @param pid
	 *            PID des DEs
	 * @return SystemObjekt Instanz
	 */
	static DeSim getDe(final String pid) {
		return SteuerModulSim.mapPidDe.get(pid);
	}

	/**
	 * Liefert alle DEs, die eine Instanz haben
	 *
	 * @return Liste der DEs
	 */
	public static Collection<DeSim> getAllDes() {
		return SteuerModulSim.mapPidDe.values();
	}

	/**
	 * Liefert die Liste aller Lve Des des Steuermoduls
	 *
	 * @return Liste der DEs
	 */
	public Collection<LveDeSim> getLveDeList() {
		return lveDeList;
	}

	/**
	 * Liefert Das System De fuer Lve Daten
	 *
	 * @return DE
	 */
	public LveDeKanal255Sim getSysLveDe() {
		return sysLveDe;
	}

	/**
	 * Liefert die Liste aller Axl Des des Steuermoduls
	 *
	 * @return Liste der DEs
	 */
	public Collection<AxlDeSim> getAxlDeList() {
		return axlDeList;
	}

	/**
	 * Liefert Das System De fuer Axl Daten
	 *
	 * @return DE
	 */
	public AxlDeKanal255Sim getSysAxlDe() {
		return sysAxlDe;
	}
}
