/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveAvailabilityListener;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Ermoeglicht eine Streambasierte ArchivAnfrage zu machen, ausserdem hoert zu,
 * ob das Archiv verfuegbar ist, falls nicht, dann macht eine Fehlermeldung und
 * stoppt die Applikation
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class StreamBasierteArchivAnfrage {

	/**
	 * Speichert den Zugriff auf das Archivsystem.
	 */
	private final ArchiveRequestManager archiveRequestManager;

	/**
	 * Die Streams, die vom Archiv als Antwort kommen
	 */
	private ArchiveDataStream[] archiveDataStreams;

	/**
	 * Debug-Logger
	 */
	public static final Debug LOGGER = Debug.getLogger();

	/**
	 * Die Resultat der Anfrage
	 */
	private boolean erfolg;

	/**
	 * Zum mitteilen wenn das Archiv offline geht
	 */
	private final ArchivVerfuegbarkeitListener verfuegbarkeitListener = new ArchivVerfuegbarkeitListener();

	/**
	 * Zeigt an, ob eine Verbindung zum Archiv besteht.
	 */
	private boolean archivVerfuegbar;

	/**
	 * Standard Konstruktor
	 *
	 * @param davVerbindung
	 *            Verbindung zum Datenverteiler
	 */
	public StreamBasierteArchivAnfrage(final ClientDavInterface davVerbindung) {
		archivVerfuegbar = davVerbindung.getArchive().isArchiveAvailable();
		archiveRequestManager = davVerbindung.getArchive();
		archiveRequestManager.addArchiveAvailabilityListener(verfuegbarkeitListener);
	}

	/**
	 * Striembasierte Anfrage zum Archiv
	 *
	 * @param objekte
	 *            Collection von System Objekten
	 * @param datenBeschreibung
	 *            DatenBeschreibung Attributgruppe, Aspekt,
	 *            Simulationsinvariante
	 * @param relativeAnfrage
	 *            Ob die anfrage Relativ ist
	 * @param intervallBeginOderAnzahl
	 *            Fals relativ, dann anzahl der Datensaetze, falls absolut, dann
	 *            Beginnzeitpunkt
	 * @param intervallEnd
	 *            Endezeitpunkt der Anfrage
	 */
	public void generiereAnfrage(final Collection<SystemObject> objekte, final DataDescription datenBeschreibung,
			final boolean relativeAnfrage, final long intervallBeginOderAnzahl, final long intervallEnd) {

		StreamBasierteArchivAnfrage.LOGGER.fine(
				getAnfrageInfo(objekte, datenBeschreibung, relativeAnfrage, intervallBeginOderAnzahl, intervallEnd));

		if (!archivVerfuegbar) {
			erfolg = false;
			StreamBasierteArchivAnfrage.LOGGER.warning(
					"Archivanfrage wird verworfen, da keine Verbindung zum Archiv besteht:\n" + getAnfrageInfo(objekte,
							datenBeschreibung, relativeAnfrage, intervallBeginOderAnzahl, intervallEnd));
		} else {
			final TimingType timingAngabe = TimingType.DATA_TIME;
			final ArchiveDataKindCombination archivDatenTypen = new ArchiveDataKindCombination(
					ArchiveDataKind.ONLINE_DELAYED, ArchiveDataKind.ONLINE); // WORKAROUND
			// ARCHIV-BUG
			final ArchiveOrder sortierung = ArchiveOrder.BY_DATA_TIME;
			final ArchiveRequestOption anfrageTyp = ArchiveRequestOption.NORMAL;
			final ArchiveQueryPriority prioritaet = ArchiveQueryPriority.MEDIUM;

			// ZeitBeschreibung
			final ArchiveTimeSpecification archivZeitBeschreibung;
			if (relativeAnfrage) {
				// WORKAROUND ARCHIV-BUG
				archivZeitBeschreibung = new ArchiveTimeSpecification(timingAngabe, false,
						intervallEnd - (4 * 3600 * 1000), intervallEnd);
			} else {
				archivZeitBeschreibung = new ArchiveTimeSpecification(timingAngabe, relativeAnfrage,
						intervallBeginOderAnzahl, intervallEnd);
			}

			final List<ArchiveDataSpecification> archivDatenBeschreibungen = new LinkedList<>();

			// eine Liste von Specifikationen fuer jedem Objekt
			for (final SystemObject objekt : objekte) {
				archivDatenBeschreibungen.add(new ArchiveDataSpecification(archivZeitBeschreibung, archivDatenTypen,
						sortierung, anfrageTyp, datenBeschreibung, objekt));
			}

			try {
				// die Anfrage
				final ArchiveDataQueryResult anfrageErgebnis = archiveRequestManager.request(prioritaet,
						archivDatenBeschreibungen);

				if (anfrageErgebnis.isRequestSuccessful()) {

					StreamBasierteArchivAnfrage.LOGGER
					.info("Archivanfrage konnte erfolgreich bearbeitet werden:\n" + getAnfrageInfo(objekte,
							datenBeschreibung, relativeAnfrage, intervallBeginOderAnzahl, intervallEnd));
					archiveDataStreams = anfrageErgebnis.getStreams();
					erfolg = true;
				} else {
					erfolg = false;
					StreamBasierteArchivAnfrage.LOGGER
					.error("Eine Archivanfrage konnte nicht bearbeitet werden:\n"
							+ getAnfrageInfo(objekte, datenBeschreibung, relativeAnfrage,
									intervallBeginOderAnzahl, intervallEnd)
							+ "\nFehler: " + anfrageErgebnis.getErrorMessage());
				}
			} catch (final IllegalStateException ex) {
				StreamBasierteArchivAnfrage.LOGGER.error("Generiere ArchivAnfrage:" + ex.getMessage());
			} catch (final InterruptedException ex) {
				StreamBasierteArchivAnfrage.LOGGER.error("Generiere ArchivAnfrage:" + ex.getMessage());
			}
		}
	}

	/**
	 * Bearbeitung der Antwort der Archivanfrage
	 *
	 * @param bearbeiter
	 *            Bearbeitet die einzelne Datensaetzte, muss das Interface
	 *            SBArchivAnfrageBearbeiter implementieren
	 */
	public void bearbeiteAntwortStream(final SBArchivAnfrageBearbeiter bearbeiter) {
		for (final ArchiveDataStream ds : archiveDataStreams) {
			try {
				ArchiveData archivDatei = ds.take();
				while (archivDatei != null) {
					if ((null != archivDatei.getData()) || DataState.END_OF_ARCHIVE.equals(archivDatei.getDataType())) { // WORKAROUND
						// ARCHIV-BUG
						bearbeiter.bearbeiteArchivDaten(archivDatei.getObject(), archivDatei.getDataDescription(),
								archivDatei.getDataType(), archivDatei.getData(), archivDatei.getDataTime());
					}
					archivDatei = ds.take();
				}

			} catch (final IllegalStateException ex) {
				StreamBasierteArchivAnfrage.LOGGER.error("Bearbeitung AntwortStream:" + ex.getMessage());
			} catch (final InterruptedException ex) {
				StreamBasierteArchivAnfrage.LOGGER.error("Bearbeitung AntwortStream:" + ex.getMessage());
			} catch (final IOException ex) {
				StreamBasierteArchivAnfrage.LOGGER.error("Bearbeitung AntwortStream:" + ex.getMessage());
			}
		}
	}

	/**
	 * Hoert zu, und reagiert, wenn die Verfuegbarkeit des Archivs sich aendert
	 */
	protected class ArchivVerfuegbarkeitListener implements ArchiveAvailabilityListener {

		/**
		 * Die Callback Methode, wenn sich die Verfuegbarkeit veraendert
		 */
		@Override
		public void archiveAvailabilityChanged(final ArchiveRequestManager archive) {
			if (!archive.isArchiveAvailable()) {
				archivVerfuegbar = false;
				StreamBasierteArchivAnfrage.LOGGER.warning("Keine Verbindung zum Archiv.");
			} else {
				archivVerfuegbar = true;
				StreamBasierteArchivAnfrage.LOGGER.config("Verbindung zum Archiv hergestellt.");
			}
		}

	}

	/**
	 * erfragt ob die Anfrage erfolgreich war
	 *
	 * @return true, wenn die Anfrabe OK ist
	 */
	public final boolean istErfolgreich() {
		return erfolg;
	}

	/**
	 * Erfragt nach dem Anzahl der Streams, die die Anfrage geliefert hat
	 *
	 * @return Anzahl der Streams
	 */
	public final int getStreamsCount() {
		return archiveDataStreams.length;
	}

	/**
	 * Liefert die Streams die man als Antwort bekommen hat
	 *
	 * @return Streams
	 */
	public final ArchiveDataStream[] getArchiveDataStreams() {
		return archiveDataStreams;
	}

	/**
	 * Liefert einen Stream den man als Antwort bekommen hat
	 *
	 * @param i
	 *            Reihenfolge des Streams
	 * @return Stream oder null, wenn es kein Stream mit Index i gibt
	 */
	public final ArchiveDataStream getArchiveDataStream(final int i) {
		if (i < archiveDataStreams.length) {
			return archiveDataStreams[i];
		} else {
			return null;
		}
	}

	/**
	 * Erfragt eine Zeichenkette die die Anfrage repraesentiert.
	 *
	 * @param objekte
	 *            Collection von System Objekten
	 * @param datenBeschreibung
	 *            DatenBeschreibung Attributgruppe, Aspekt,
	 *            Simulationsinvariante
	 * @param relativeAnfrage
	 *            Ob die anfrage Relativ ist
	 * @param intervallBeginOderAnzahl
	 *            Fals relativ, dann anzahl der Datensaetze, falls absolut, dann
	 *            Beginnzeitpunkt
	 * @param intervallEnd
	 *            Endezeitpunkt der Anfrage
	 */
	private String getAnfrageInfo(final Collection<SystemObject> objekte, final DataDescription datenBeschreibung,
			final boolean relativeAnfrage, final long intervallBeginOderAnzahl, final long intervallEnd) {
		String s = "Anfrage (" + (relativeAnfrage ? "relativ" : "absolut") + "):\n";

		if (objekte == null) {
			s += "  Objekte ist <<null>>\n";
		} else {
			if (objekte.isEmpty()) {
				s += "  keine Objekte\n";
			} else {
				s += "  Objekte:\n";
				for (final SystemObject obj : objekte) {
					s += "    " + obj.toString() + "\n";
				}
				if (datenBeschreibung != null) {
					s += "  Datenbeschreibung (atg, asp, sim): ("
							+ (datenBeschreibung.getAttributeGroup() == null ? "<<null>>"
									: datenBeschreibung.getAttributeGroup().getPid())
							+ ", "
							+ (datenBeschreibung.getAspect() == null ? "<<null>>"
									: datenBeschreibung.getAspect().getPid())
							+ ", " + datenBeschreibung.getSimulationVariant() + ")\n";
				} else {
					s += "  Datenbeschreibung ist <<null>>\n";
				}
				if (relativeAnfrage) {
					s += "  Anzahl: " + intervallBeginOderAnzahl + "\n";
					s += "  Intervallende: " + intervallEnd + "\n";
				} else {
					final SimpleDateFormat dateFormat = new SimpleDateFormat(DUAKonstanten.ZEIT_FORMAT_GENAU_STR);
					s += "  Abfrageintervall: [" + dateFormat.format(new Date(intervallBeginOderAnzahl)) + ", "
							+ dateFormat.format(new Date(intervallEnd)) + "]\n";
				}
			}

		}

		return s;
	}

}
